@@ -23,10 +23,9 @@
 #'
 #' @description Given the values of (observed) x-variables in a SEM,
 #' this function may be used to predict the values of (observed) y-variables.
-#' The predictive procedure consists of two steps: (1) construction of the
-#' topological layer (TL) ordering of the input graph; (2) prediction of
-#' the node y values in a layer, where the nodes included in the previous
-#' layers act as predictors x. 
+#' The predictive procedure consists of two steps. First, the topological layer
+#' ordering of the input graph is defined. Then, the node y values in a layer are
+#' predicted, where the nodes in successive layers act as x-predictors. 
 #' 
 #' @param object An object, as that created by the function \code{SEMrun()}
 #' with the argument \code{group} set to the default \code{group = NULL}.
@@ -68,13 +67,15 @@
 #' 
 #' de Rooij M, Karch JD, Fokkema M, Bakk Z, Pratiwi BC, and Kelderman H
 #' (2023). SEM-Based Out-of-Sample Predictions, Structural Equation Modeling:
-#' A Multidisciplinary Journal, 30:1, 132-148
-#' <https://doi.org/10.1080/10705511.2022.2061494>
+#' A Multidisciplinary Journal, 30:1, 132-148. <https://doi.org/10.1080/10705511.2022.2061494>
 #'
 #' Grassi M, Palluzzi F, Tarantino B (2022). SEMgraph: An R Package for Causal Network
 #' Analysis of High-Throughput Data with Structural Equation Models.
-#' Bioinformatics, 38 (20), 4829–4830 <https://doi.org/10.1093/bioinformatics/btac567>
-#' 
+#' Bioinformatics, 38 (20), 4829–4830. <https://doi.org/10.1093/bioinformatics/btac567>
+#'
+#' Grassi, M., Tarantino, B. (2025). SEMdag: Fast learning of Directed Acyclic Graphs via
+#' node or layer ordering. PLoS ONE 20(1): e0317283. https://doi.org/10.1371/journal.pone.0317283
+#'
 #' @examples
 #'
 #' # load ALS data
@@ -170,7 +171,8 @@ predict.SEM <- function(object, newdata, newoutcome = NULL, verbose = FALSE, ...
 	}
 
 	PE<- colMeans((yobs - yhat)^2)
-	PE<- ifelse(PE > 1, NA, PE)
+	#PE <- ifelse(PE > 1, NA, PE)
+	PE <- ifelse(PE > 1, 1, PE)
 	pe<- mean(PE, na.rm = TRUE)
 			
 	Shat<- Sigma
@@ -186,13 +188,15 @@ predict.SEM <- function(object, newdata, newoutcome = NULL, verbose = FALSE, ...
 
 #' @title Create a plot for a neural network model
 #'
-#' @description The function draws a neural network plot as a neural
-#' interpretation diagram using with the \code{\link[NeuralNetTools]{plotnet}}
-#' function of the \pkg{NeuralNetTools} R package.
+#' @description The function uses the \code{\link[NeuralNetTools]{plotnet}}
+#' function of the \pkg{NeuralNetTools} R package to draw a neural network
+#' plot and visualize the hidden layer structure. 
 #'
-#' @param dnn.fit A neural network model from \pkg{cito} R package. 
+#' @param object A neural network model object
+#' @param hidden The hidden structure of the object
 #' @param bias A logical value, indicating whether to draw biases in 
-#' the layers (default = FALSE). 
+#' the layers (default = FALSE).
+#' @param sleep Suspend plot display for a specified time (in secs, default = 2).
 #' @param ... Currently ignored.
 #'
 #' @details The induced subgraph of the input graph mapped on data 
@@ -200,7 +204,7 @@ predict.SEM <- function(object, newdata, newoutcome = NULL, verbose = FALSE, ...
 #' weight W > 0, the connection is activated and it is highlighted in red;  
 #' if W < 0, the connection is inhibited and it is highlighted in blue.  
 #' 
-#' @return The function invisibly returns the graphical object representing
+#' @return The function invisibly returns the graphical objects representing
 #' the neural network architecture designed by NeuralNetTools. 
 #' 
 #' @author Mario Grassi \email{mario.grassi@unipv.it}
@@ -221,48 +225,56 @@ predict.SEM <- function(object, newdata, newoutcome = NULL, verbose = FALSE, ...
 #' data<- transformData(data)$data
 #'
 #' #ncores<- parallel::detectCores(logical = FALSE)
-#' dnn0 <- SEMdnn(ig, data, train=1:nrow(data), grad = FALSE,
-#' 			#hidden = 5*K, link = "selu", bias = TRUE,
+#' dnn0 <- SEMdnn(ig, data, train=1:nrow(data), algo = "layerwise",
 #' 			hidden = c(10, 10, 10), link = "selu", bias =TRUE,
-#' 			validation = 0, epochs = 32, ncores = 2)
+#' 			epochs = 32, patience = 10, verbose = TRUE)
 #'
-#'  for (j in 1:length(dnn0$model)) {
-#'    nnj <- dnn0$model[[j]][[1]]
-#'    nplot(nnj)
-#'    Sys.sleep(5)
-#'  }
+#'  #Visualize the neural networks per each layer of dnn0
+#'  nplot(dnn0, hidden = c(10, 10, 10), bias = FALSE)
 #' }
 #' }
 #'
 #' @export
 #' 
-nplot<- function(dnn.fit, bias=FALSE, ...)
+nplot <- function (object, hidden, bias = TRUE, sleep = 2, ...) 
 {
-	W<- coef(dnn.fit)[[1]] #str(W)
-	w0<- list()
-	for (i in seq(2, length(W), by=2)){
-	  wi<- cbind(W[[i]],W[[i-1]])
-	  w0<- c(w0, list(t(wi)))
-	} #w0
-	w1<- unlist(w0) #w1
-	input<- dnn.fit$ model_properties$input
-	hidden<- dnn.fit$ model_properties$hidden
-	output<- dnn.fit$ model_properties$output
-	struct1<- c(input, hidden, output)
-	g<- NeuralNetTools::plotnet(w1, struct1, pos_col="red", neg_col="blue", bias=bias)
+    stopifnot(inherits(object$model[[1]][[1]], "DNN"))
+	algo <- object$model[[1]][[1]]$param$algo
+	#if (algo == "autoencoder") stop("algo = \"", algo, "\" is not supported!")
+	model <- object$model[[1]]
+	g <- list()
 	
-	return(invisible(g))
+    for (j in 1:length(model)) {
+        W <- model[[j]]$weights[[1]]
+        input <- ncol(W[[1]])
+        if (algo == "structured") {
+		 if (hidden[1] < input) hidden <- hidden + input #n_hid > n_inp 
+		}
+		output <- nrow(W[[length(W) - 1]])
+        w0 <- list()
+        for (i in seq(2, length(W), by = 2)) {
+            wi <- cbind(W[[i]], W[[i - 1]])
+            w0 <- c(w0, list(t(wi)))
+        }
+        w1 <- unlist(w0)
+        struct1 <- c(input, hidden, output)
+        g[[j]] <- NeuralNetTools::plotnet(w1, struct1, pos_col = "red", 
+            neg_col = "blue", bias = bias)
+        Sys.sleep(sleep)
+    }
+	
+    return(invisible(g))
 }
 
 #' @title Map additional variables (nodes) to a graph object
 #'
 #' @description The function insert additional nodes to a graph object.
 #' Among the node types, additional source or sink nodes can be added. 
-#' Regarding the former, source nodes can represent: (i) data variables; 
-#' (ii) a group variable; (iii) Latent Variables (LV). For the latter,
-#' sink nodes represent the levels of a categorical outcome variable and
-#' are linked with all graph nodes.' Moreover, \code{mapGraph()} can also
-#' create a new graph object starting from a compact symbolic formula. 
+#' Source nodes can represent: (i) data variables; (ii) a group variable;
+#' (iii) latent variables (LV). Vice versa, sink nodes represent the levels
+#' of a categorical outcome variable and are linked with all graph nodes.
+#' Moreover, \code{mapGraph()} can also create a new graph object starting
+#' from a compact symbolic formula. 
 #' 
 #' @param graph An igraph object.
 #' @param type A character value specifying the type of mapping. Five 
@@ -466,7 +478,7 @@ buildLevels <- function(dag, ...)
 	l1<- Leaf_removal(dag)
 	if (length(l1) == 2) return(l1)
 	# leaf removal(dagT)
-	adj <- as_adj(dag, sparse=FALSE)
+	adj <- as_adjacency_matrix(dag, sparse=FALSE)
 	dagT <- graph_from_adjacency_matrix(t(adj), mode="directed")
 	l2 <- Leaf_removal(dagT)
 	l2 <- rev(l2)
@@ -746,10 +758,10 @@ dplot<- function(yobs, yhat, thr, rec, ...)
 #' res1 <- SEMml(ig, data, algo="tree")
 #' res2 <- SEMml(ig, data, algo="rf")
 #' res3 <- SEMml(ig, data, algo="xgb")
-#' res4 <- SEMml(ig, data, algo="nn")
+#' res4 <- SEMml(ig, data, algo="sem")
 #' 
 #' models <- list(res1,res2,res3,res4)
-#' names(models) <- c("tree","rf","xgb","nn")
+#' names(models) <- c("tree","rf","xgb","sem")
 #' 
 #' res.cv1 <- crossValidation(models, outcome=NULL, K=5, R=10)
 #' print(res.cv1$stats)
@@ -855,82 +867,80 @@ crossValidationR1 <- function(models, outcome, K, ncores, seed = NULL, ...)
 	if (is.null(seed)) seed <- runif(1,0,1000)
 	set.seed(seed)
 	#create folds
-	C<- length(levels(outcome))
-	n<- nrow(models[[1]]$data)
+	C <- length(levels(outcome))
+	n <- nrow(models[[1]]$data)
 	if (C == 0) {
-	  idx<- createFolds(y=rnorm(n), k=K) #N.B. list of test indices!
+	  idx <- createFolds(y=rnorm(n), k=K) #N.B. list of test indices!
 	}else{
-	  idx<-  createFolds(y=outcome, k=K) #N.B. list of test indices!
+	  idx <-  createFolds(y=outcome, k=K) #N.B. list of test indices!
 	}
 
 	   # Initialise M-models loop 
-	   resm<- list()
-		for (m in 1:length(models)) {
+	   resm <- list()
+		for (m in 1:length(models)) { #m=1
 		 #message(names(models)[m])
-		 object<- models[[m]]  # str(models[[m]], max.level=1)
-		 graph<- models[[m]]$graph
-		 data<- models[[m]]$data
+		 object <- models[[m]]  # str(models[[m]], max.level=1)
+		 graph <- models[[m]]$graph
+		 data <- models[[m]]$data
 		 if (!is.null(outcome)) {
-		  out<- model.matrix(~outcome-1)
-		  colnames(out)<- gsub("outcome", "", colnames(out))
-		  data<- cbind(out, data)
-		  graph<- mapGraph(graph, type="outcome", C=ncol(out))
+		  out <- model.matrix(~outcome-1)
+		  colnames(out) <- gsub("outcome", "", colnames(out))
+		  data <- cbind(out, data)
+		  graph <- mapGraph(graph, type="outcome", C=ncol(out))
 		  V(graph)$name[igraph::degree(graph, mode="out") == 0]<- colnames(out)
 		 }
 
 		 # Initialise K-fold cross-validation loop
-		  pb <- pb(nrep = K, snow = FALSE)
+		  pb0 <- pb(nrep = K, snow = FALSE)
 		  resk <- list()
 			for (k in 1:K) {
-			  pb$tick()
-			  train<- c(1:nrow(data))[-idx[[k]]]
+			  pb0$tick()
+			  train <- c(1:nrow(data))[-idx[[k]]]
 			  if (inherits(object, "SEM")) {
-			   res0<- quiet(SEMgraph::SEMrun(graph, data[train, ], algo="ricf", n_rep=0))
+			   res0 <- quiet(SEMgraph::SEMrun(graph, data[train, ], algo="ricf", n_rep=0))
 			  }
 			  if (inherits(object, "ML")) {
-			   ml0<- c("SEM", "rpart", "ranger", "xgb.Booster", "nnet", "citodnn")
-			   ml1<- c("sem", "tree", "rf", "xgb", "nn", "dnn")
-			  for (ml in 1:6){
+			   ml0 <- c("SEM", "rpart", "ranger", "xgb.Booster")
+			   ml1 <- c("sem", "tree", "rf", "xgb")
+			  for (ml in 1:4){
 			   if (inherits(object$model[[1]][[1]], ml0[ml])) algo<- ml1[ml]
 			  }
 			   res0<- quiet(SEMdeep::SEMml(graph, data[train, ], algo = algo))
 			  }
-			  if (inherits(object, "DNN")) {
-			   loss<- object$model[[1]][[1]]$loss$call
-			   device<- object$model[[1]][[1]]$device
-			   mp<- object$model[[1]][[1]]$model_properties
-			   mt<- object$model[[1]][[1]]$training_properties
-
-			   res0<- quiet(SEMdeep::SEMdnn(graph, data[train, ], thr = NULL, 
-				hidden = mp[[3]], link = mp[[4]], bias = mp[[5]], dropout = mp[[6]],
-				lr = mt[[1]], optimizer = mt[[3]], epochs = mt[[4]], early_stopping = mt[[5]],
-				loss = loss, validation = mt[[7]], lambda = mt[[8]], alpha = mt[[9]],
-				device = device, ncores = ncores, verbose = FALSE))		
+			  if (inherits(object, "DNN")) { #str(object, max.level=1)
+			   mp <- object$model[[1]][[1]]$param
+			   mp <- object$model[[1]][[1]]$param
+			   res0 <- quiet(SEMdeep::SEMdnn(graph, data[train, ], algo = mp[[1]],
+				hidden = mp[[2]], link = mp[[3]], bias = mp[[4]], dropout = mp[[5]],
+				loss = "mse", validation = mp[[7]], lambda = mp[[8]], alpha = mp[[9]],
+				optimizer = "adam", lr = mp[[11]], batchsize = mp[[12]], shuffle = mp[[13]],
+				baseloss = mp[[14]], burnin = mp[[15]], thr = NULL, nboot = 0,
+				epochs = mp[[18]], patience = mp[[19]], device = "cpu", verbose = FALSE))
 			  }
-			  yhat0<- predict(res0, data[-train, ])# dim(yhat0$Yhat)
+			  yhat0 <- predict(res0, data[-train, ])# dim(yhat0$Yhat)
 
 			  if (is.null(outcome)) {
-			   resk[[k]]<- data.frame(amse=yhat0$PE[1],r2=yhat0$PE[2],srmr=yhat0$PE[3])
+			   resk[[k]] <- data.frame(amse=yhat0$PE[1],r2=yhat0$PE[2],srmr=yhat0$PE[3])
 			  }else{
-			   yobs<- outcome[-train]
-			   yhat<- yhat0$Yhat[ ,levels(outcome)]
-			   resk[[k]]<- SEMdeep::classificationReport(yobs, yhat)$CM
+			   yobs <- outcome[-train]
+			   yhat <- yhat0$Yhat[ ,levels(outcome)]
+			   resk[[k]] <- SEMdeep::classificationReport(yobs, yhat)$CM
 			  }
 			} # end K-folds loop
 
 			if (is.null(outcome)) {
-			 RESK<-  do.call(rbind, lapply(resk, as.data.frame))
+			 RESK <-  do.call(rbind, lapply(resk, as.data.frame))
 			}else{
-			 CMr<- base::Reduce('+', resk)
-			 RESK<- SEMdeep::classificationReport(CM=CMr)$stats[C+2,c(3:5)]
+			 CMr <- base::Reduce('+', resk)
+			 RESK <- SEMdeep::classificationReport(CM=CMr)$stats[C+2,c(3:5)]
 			}
-		  resm[[m]]<- apply(RESK, 2, mean)
+		  resm[[m]] <- apply(RESK, 2, mean)
 		  message(" ", K, "-fold ", names(models)[m], " done.")
 		} # end M-models loop
 
 		RESM<- do.call(cbind, lapply(resm, matrix))
-		rownames(RESM)<- names(resm[[1]])
-		colnames(RESM)<- names(models)
+		rownames(RESM) <- names(resm[[1]])
+		colnames(RESM) <- names(models)
 
 	return(list(stats = t(RESM), PE = NULL))
 }
