@@ -35,15 +35,13 @@
 #' @param outcome A character vector (as.fctor) of labels for a categorical
 #' output (target). If NULL (default), the categorical output (target) will
 #' not be considered.
-#' @param algo ML method used for nodewise-network predictions.
-#' Six algorithms can be specified:
+#' @param algo ML method used for nodewise-based predictions.
+#' Four algorithms can be specified:
 #' \itemize{
 #' \item \code{algo="sem"} (default) for a linear SEM, see \code{\link[SEMgraph]{SEMrun}}. 
 #' \item \code{algo="tree"} for a CART model, see \code{\link[rpart]{rpart}}.
 #' \item \code{algo="rf"} for a random forest model, see \code{\link[ranger]{ranger}}.
 #' \item \code{algo="xgb"} for a XGBoost model, see \code{\link[xgboost]{xgboost}}.
-#' \item \code{algo="nn"} for a small neural network model (1 hidden layer and 10 nodes), see \code{\link[nnet]{nnet}}.
-#' \item \code{algo="dnn"} for a large neural network model (1 hidden layers and 1000 nodes), see \code{\link[cito]{dnn}}.
 #' }
 #' @param thr A numeric value [0-1] indicating the threshold to apply to the
 #' variable importance values to color the graph. If thr = NULL (default), the
@@ -56,21 +54,20 @@
 #' @param ... Currently ignored.
 #'
 #' @details By mapping data onto the input graph, \code{SEMml()} creates
-#' a set of nodewise-based models based on the directed links, i.e., 
+#' a set of nodewise models based on the directed links, i.e., 
 #' a set of edges pointing in the same direction, between two nodes 
 #' in the input graph that are causally relevant to each other. 
-#' The mediator or sink variables can be characterized in detail as 
-#' functions of their parents. An ML model (sem, tree, rf, xgb, nn, dnn) 
-#' can then be fitted to each variable with non-zero inbound connectivity. 
-#' With R representing the number of mediators and sink nodes in the 
-#' network, the model fitting process is performed equation-by-equation 
-#' (r=1,...,R) times.
+#' The mediator or sink variables are defined as functions of their parents.
+#' Then, an ML model (sem, tree, rf, xgb) can be fitted to each variable with
+#' non-zero inbound connectivity. The model fitting process is performed
+#' equation-by-equation (r=1,...,R) times, where R represents the number of
+#' mediators and sink nodes in the input graph.
 #'
 #' If boot != 0, the function will implement the cheap bootstrapping proposed by
-#' Lam (2002) to generate uncertainties, i.e. 90% confidence intervals (90% CIs),
-#' for ML parameters. Bootstrapping can be enabled by setting a small number
-#' (1 to 10) of bootstrap samples. Note, however, that the computation can be
-#' time-consuming for massive MLs, even with cheap bootstrapping!
+#' Lam (2002) to generate uncertainties (i.e., bootstrap \code{90\%CIs}) for ML
+#' parameters. Bootstrapping can be enabled by setting a small number (1 to 10) of
+#' bootstrap samples. Note, however, that the computation can be time-consuming for
+#' massive MLs, even with cheap bootstrapping!
 #'
 #' @return An S3 object of class "ML" is returned. It is a list of 5 objects:
 #' \enumerate{
@@ -89,7 +86,7 @@
 #' will be highlighted in pink.
 #' \item "data", input data subset mapping graph nodes.
 #' }
-#' Using the default \code{algo="sem"}, the usual output of a linear nodewise-based,
+#' Using the default \code{algo="sem"}, the usual output of a linear nodewise-based
 #' SEM, see \code{\link[SEMgraph]{SEMrun}} (algo="cggm"), will be returned. 
 #'
 #' @author Mario Grassi \email{mario.grassi@unipv.it}
@@ -109,10 +106,8 @@
 #' Proceedings of the 22nd ACM SIGKDD International Conference on Knowledge 
 #' Discovery and Data Mining.
 #' 
-#' Ripley B.D. (1996). Pattern Recognition and Neural Networks. Cambridge University Press.
-#' 
 #' Lam, H. (2022). Cheap bootstrap for input uncertainty quantification. WSC '22:
-#' Proceedings of the Winter Simulation Conference, 2318 - 2329.
+#' Proceedings of the Winter Simulation Conference, 2318-2329.
 #'
 #' @examples
 #'
@@ -137,38 +132,38 @@
 #' # ... xgb
 #' res3<- SEMml(ig, data[train, ], algo="xgb")
 #'
-#' # ... nn
-#' res4<- SEMml(ig, data[train, ], algo="nn")
+#' # ... sem
+#' res4<- SEMml(ig, data[train, ], algo="sem")
 #' 
 #' end<- Sys.time()
 #' print(end-start)
 #'
-#' #visualizaation of the colored dag for algo="nn"
-#' gplot(res4$graph, l="dot", main="nn")
+#' #visualizaation of the colored dag for algo="sem"
+#' gplot(res4$graph, l="dot", main="sem")
 #'
 #' #Comparison of fitting indices (in train data)
 #' res1$fit$fitIdx #tree
 #' res2$fit$fitIdx #rf
 #' res3$fit$fitIdx #xgb
-#' res4$fit$fitIdx #nn
+#' res4$fit$fitIdx #sem
 #' 
 #' #Comparison of parameter estimates (in train data)
 #' parameterEstimates(res1$fit) #tree
 #' parameterEstimates(res2$fit) #rf
 #' parameterEstimates(res3$fit) #xgb
-#' parameterEstimates(res4$fit) #nn
+#' parameterEstimates(res4$fit) #sem
 #' 
 #' #Comparison of VarImp (in train data)
 #' table(E(res1$graph)$color) #tree
 #' table(E(res2$graph)$color) #rf
 #' table(E(res3$graph)$color) #xgb
-#' table(E(res4$graph)$color) #nn
+#' table(E(res4$graph)$color) #sem
 #'
 #' #Comparison of AMSE, R2, SRMR (in test data)
 #' print(predict(res1, data[-train, ])$PE) #tree
 #' print(predict(res2, data[-train, ])$PE) #rf
 #' print(predict(res3, data[-train, ])$PE) #xgb
-#' print(predict(res4, data[-train, ])$PE) #nn
+#' print(predict(res4, data[-train, ])$PE) #sem
 #' 
 #' #...with a categorical (as.factor) outcome
 #' outcome <- factor(ifelse(group == 0, "control", "case")); table(outcome) 
@@ -209,7 +204,7 @@ SEMml <- function(graph, data, outcome = NULL, algo = "sem", thr = NULL, nboot=0
 	# extract parameter estimates:
 	res <- list()
 	nboot <- nboot + 1
-	pb <- pb(nrep = nboot, snow = FALSE)
+	pb0 <- pb(nrep = nboot, snow = FALSE)
 	message("Running SEM model via ML...")
 	
 	  for (b in 1:nboot) {
@@ -222,7 +217,7 @@ SEMml <- function(graph, data, outcome = NULL, algo = "sem", thr = NULL, nboot=0
 		}else{
 		 res[[b]] <- parameterEstimates.ML(dag, data = data[idx, ], algo, nboot, ncores)
 		}
-		pb$tick()
+		pb0$tick()
 		message(" done.")
 	   #print(all.vars(res[[b]]$dnn[[1]]$old_formula));cat("\n")
 	  }
@@ -283,72 +278,30 @@ parameterEstimates.ML <- function(dag, data, algo, nboot, ncores, ...)
 	f <- list()
 	vx <- list()
 	n <- nrow(Z_train)
-	
+
 	for( j in 1:length(y)){
 	 vx[[j]] <- y[[j]][,1]
-	  # Fit a linear L1 model of Y on X	
 	  if (length(vx[[j]]) >= n) {
-	   if (nboot == 1) {
-	  	X <- as.matrix(Z_train[ ,vx[[j]]])
-	  	Y <- as.numeric(Z_train[ ,names(y)[j]])
-	  	fit <- glmnet::cv.glmnet(X, Y, family="gaussian")
-	  	B <- coef(fit, s = "lambda.min")[-1, ]
-	  	vx[[j]] <- vx[[j]][abs(B) > 0.05]
-	   }else{
-	    #set.seed(runif(1,0,1000))
-		vx[[j]] <- vx[[j]][sample(1:length(vx[[j]]), 0.5*n)]
-	   }
+	    set.seed(1)
+		vx[[j]] <- vx[[j]][sample(1:length(vx[[j]]), round(0.9*n))]
 	  }
  	 C <- paste(vx[[j]], collapse = " + ")
 	 f[[j]] <- formula(paste(names(y)[j], "~", C))
 	}
-		
-	#fitting a dnn model to predict Y on X
+
+	#fitting a ML model to predict Y on X
 	#pos <- 1
 	envir <- as.environment(1)
 	assign('f', f, envir = envir)
 	nrep <- length(y)
 
-	if (algo == "dnn") {
-	 #require(foreach)
-	 cl <- parallel::makeCluster(ncores)
-	 doSNOW::registerDoSNOW(cl)
-	 opts <- list(progress = pb(nrep = nrep, snow = TRUE))
-	 #pb <- txtProgressBar(min=0, max=nrep, style=3)
-	 #progress <- function(n) setTxtProgressBar(pb, n)
-	 #opts <- list(progress = progress)
-
-	 fit<- foreach(j=1:nrep, .options.snow=opts) %dopar% {
-	  ml.fit <- cito::dnn(eval(f[[j]]),
-						#as.formula(f),
-						data = Z_train, 
-						hidden = 1000,
-						activation = "selu",
-						bias = TRUE,
-						dropout = 0,
-						loss = "mse",
-						validation = 0,
-						lambda = 0,
-						alpha = 0.5,
-						optimizer = "adam",
-						lr = 0.01,
-						epochs = 100,
-						plot = FALSE,
-						verbose = FALSE,
-						device = "cpu",
-						early_stopping = FALSE)
-	 }
-	 #close(pb)
-	 parallel::stopCluster(cl)
-	}
-
 	#opb<- pbapply::pboptions(type = "timer", style = 2)
-	pb <- pb(nrep = nrep, snow = FALSE)
+	pb0 <- pb(nrep = nrep, snow = FALSE)
 
 	if (algo == "tree") {
 	 #fit <- pbapply::pblapply(1:nrep, function(x){
 	 fit <- lapply(1:nrep, function(x){
-	  pb$tick()
+	  pb0$tick()
 	  cp<- rpart::rpart.control(cp = 0.01) #default cp
 	  ml.fit <- rpart::rpart(eval(f[[x]]),
 						data = data.frame(Z_train),
@@ -360,12 +313,12 @@ parameterEstimates.ML <- function(dag, data, algo, nboot, ncores, ...)
 	if (algo == "rf") {
 	 #fit <- pbapply::pblapply(1:nrep, function(x){
 	 fit <- lapply(1:nrep, function(x){
-	   pb$tick()
+	   pb0$tick()
 	   ml.fit <- ranger::ranger(eval(f[[x]]),
 						data = Z_train,
 						num.trees = 500, #default
 						mtry = NULL,     #default
-						importance = "impurity", #"none"
+						importance = "permutation", #"impurity"
 						num.threads = ncores) #2
 	 })
 	 #}, cl=NULL)
@@ -374,7 +327,7 @@ parameterEstimates.ML <- function(dag, data, algo, nboot, ncores, ...)
 	if (algo == "xgb") {
 	 #fit <- pbapply::pblapply(1:nrep, function(x){
 	 fit <- lapply(1:nrep, function(x){
-	  pb$tick()
+	  pb0$tick()
 	  X <- vx[[x]]
 	  Y <- names(y)[x]
 	  ml.fit <- xgboost::xgboost(data = as.matrix(Z_train[, X]),
@@ -390,24 +343,6 @@ parameterEstimates.ML <- function(dag, data, algo, nboot, ncores, ...)
 	 #}, cl=NULL)												
 	}
 
-	if (algo == "nn") {
-	 #fit <- pbapply::pblapply(1:nrep, function(x){
-	 fit <- lapply(1:nrep, function(x){
-	  pb$tick()
-	  ml.fit <- nnet::nnet(eval(f[[x]]),
-						data = Z_train,
-						linout = TRUE,#FALSE default
-						size = 10,
-						rang = 0.3,#0.7 default
-						decay = 5e-4,#0 default
-						maxit = 1000,#100 default
-						trace = FALSE,#default
-						MaxNWts = 5000)#1000 default
-	 })
-	 #}, cl=NULL)
-	}
-	#str(fit, max.level=1)
-
 	sigma<- NULL
 	YHAT<- NULL
 	for (j in 1:length(y)) {
@@ -433,7 +368,7 @@ parameterEstimates.ML <- function(dag, data, algo, nboot, ncores, ...)
 	return(list(ml = fit, sigma = sigma, YHAT = YHAT))
 }
 
-#' @title SEM-based out-of-sample prediction using node-wise ML
+#' @title SEM-based out-of-sample prediction using nodewise ML
 #'
 #' @description Predict method for ML objects.
 #'
@@ -485,8 +420,8 @@ parameterEstimates.ML <- function(dag, data, algo, nboot, ncores, ...)
 #' res3<- SEMml(ig, data[train, ], algo="xgb")
 #' mse3<- predict(res3, data[-train, ], verbose=TRUE)
 #' 
-#' # ... nn
-#' res4<- SEMml(ig, data[train, ], algo="nn")
+#' # ... sem
+#' res4<- SEMml(ig, data[train, ], algo="sem")
 #' mse4<- predict(res4, data[-train, ], verbose=TRUE)
 #' end<- Sys.time()
 #' print(end-start)
@@ -545,7 +480,8 @@ predict.ML <- function(object, newdata, newoutcome=NULL, ncores=2, verbose=FALSE
 
 	yobs <- Z_test[, yn]
 	PE <- colMeans((yobs - yhat)^2)
-	PE <- ifelse(PE > 1, NA, PE)
+	#PE <- ifelse(PE > 1, NA, PE)
+	PE <- ifelse(PE > 1, 1, PE)
 	pe <- mean(PE, na.rm = TRUE)
 	colnames(yhat) <- sub(".", "", yn)
 	names(PE) <- colnames(yhat)
@@ -575,10 +511,9 @@ predict.ML <- function(object, newdata, newoutcome=NULL, ncores=2, verbose=FALSE
 #'
 #' @details The variable (predictor) importance will be computed considering:
 #' (i) the absolute value of the z-statistic of the model parameters for "sem";
-#' (ii) the variable importance measures from the \code{\link[rpart]{rpart}},
+#' and (ii) the variable importance measures from the \code{\link[rpart]{rpart}},
 #' \code{\link[ranger]{importance}} or \code{\link[xgboost]{xgb.importance}} functions
-#' for "tree", "rf" or "xgb"; and (iii) the Olden's connection weights for "nn" or "dnn"
-#' methods.
+#' for "tree", "rf" or "xgb" methods.
 #'
 #' @return A list of three object: (i) est: a data.frame including the connections together
 #' with their variable importances (VarImp)), (ii) gest: if the outcome vector is given,
@@ -612,9 +547,11 @@ predict.ML <- function(object, newdata, newoutcome=NULL, ncores=2, verbose=FALSE
 #' 
 getVariableImportance <- function(object, thr = NULL, verbose = FALSE, ...)
 {
-	# Set ML objects
-	ml0<- c("SEM","rpart","ranger","xgb.Booster","nnet","citodnn")
+	#stopifnot(inherits(object, c("SEM", "ML")))
+	ml0 <- c("SEM", "rpart", "ranger", "xgb.Booster")
 	stopifnot(inherits(object$model[[1]][[1]], ml0))
+		
+	# Set ML objects
 	dag <- object$graph #gplot(dag)
 	out <- unique(object$gest$lhs)
 	if (length(out) > 0) {
@@ -624,12 +561,12 @@ getVariableImportance <- function(object, thr = NULL, verbose = FALSE, ...)
 	
 	nboot<- length(object$model)
 	est_boot <- list()
-	#pb <- SEMdeep:::pb(nrep = nboot , snow = FALSE)
+	pb0 <- pb(nrep = nboot , snow = FALSE)
 	
 	for(b in 1:nboot) { #b=1
-	  #pb$tick()
+	  pb0$tick()
 	   
-	  fit<- object$model[[b]]
+	  fit <- object$model[[b]]
 	  formula <- fit$old_formula
 	  nrep <- length(fit) - 1
 	  est <- NULL
@@ -663,22 +600,6 @@ getVariableImportance <- function(object, thr = NULL, verbose = FALSE, ...)
 			vim <- as.numeric(vim[X])
 			}
 		 }
-		 if (inherits(fit[[1]], "nnet")) {
-			p <- length(fit[[j]]$coefnames)
-			w <- coef(fit[[j]])
-			Wx <- matrix(w[-c((length(w)-10):length(w))], nrow=p+1, ncol=10)
-			wy <- matrix(w[c((length(w)-10):length(w))])
-			vim <- as.numeric(Wx[-1,] %*% wy[-1,])
-		 }
-		 if (inherits(fit[[1]], "citodnn")) {
-			W <- coef(fit[[j]])[[1]]
-			w <- t(W[[1]]) %*% t(W[[3]])
-			#for (k in seq(3, length(W), by=2)) {
-			# w <- w %*% t(W[[k]])
-			# if (length(W) == 4) break
-			#}
-			vim <- as.numeric(w)
-		 }
 
 		 if (length(vim) == 0) vim<- rep(NA, length(X))
 		  estj <- data.frame(
@@ -704,16 +625,13 @@ getVariableImportance <- function(object, thr = NULL, verbose = FALSE, ...)
 
 #' @title Compute variable importance using Shapley (R2) values
 #'
-#' @description This function computes variable contributions for individual
-#' predictions using the Shapley values, a method from cooperative game
-#' theory where the variable values of an observation work together to achieve
-#' the prediction. In addition, to make variable contributions easily explainable, 
-#' the function decomposes the entire model R-Squared (R2 or the coefficient
-#' of determination) into variable-level attributions of the variance
-#' (Redell, 2019).
+#' @description This function computes a model-agnostic variable importance
+#' based on a Shapley-value decomposition of the model R-Squared (R2, i.e., the
+#' coefficient of determination) that allocates the proportion of model-
+#' explained variability in the data to each model feature (Redell, 2019).
 #'
-#' @param object A model fitting object from \code{SEMml()}, \code{SEMdnn()} or
-#' \code{SEMrun()} functions.
+#' @param object A model fitting object from \code{SEMml()}, or \code{SEMrun()} 
+#' functions.
 #' @param newdata A matrix containing new data with rows corresponding to subjects,
 #' and columns to variables.
 #' @param newoutcome A new character vector (as.factor) of labels for a categorical
@@ -726,33 +644,25 @@ getVariableImportance <- function(object, thr = NULL, verbose = FALSE, ...)
 #' graph will not be plotted to screen.
 #' @param ... Currently ignored.
 #'
-#' @details Lundberg & Lee (2017) proposed a unified approach to both
-#' local explainability (the variable contribution of a single variable within
-#' a single sample) and global explainability (the variable contribution of the
-#' entire model) by applying the fair distribution of payoffs principles from
-#' game theory put forth by Shapley (1953). Now called SHAP (SHapley Additive
-#' exPlanations), this suggested framework explains predictions of ML models,
-#' where input variables take the place of players, and their contribution to
-#' a particular prediction is measured using Shapley values. 
-#' Successively, Redell (2019) presented a metric that combines the additive 
-#' property of Shapley values with the robustness of the  R-squared (R2) of
-#' Gelman (2018) to produce a variance decomposition that accurately captures
-#' the contribution of each variable to the explanatory power of the model. 
-#' We also use the signed R2, in order to denote the regulation of connections
-#' in line with a linear SEM, since the edges in the DAG indicate node regulation
-#' (activation, if positive; inhibition, if negative). This has been recovered for
-#' each edge using sign(beta), i.e., the sign of the coefficient estimates from a
-#' linear model (lm) fitting of the output node on the input nodes, as suggested
-#' by Joseph (2019).
-#' Additionally, in order to ascertain the local significance of node regulation
-#' with respect to the DAG, the Shapley decomposition of the R-squared values for
-#' each outcome node (r=1,...,R) can be computed by summing the ShapleyR2 indices
-#' of their input nodes.
-#' Finally, It should be noted that the operations required to compute kernel SHAP
-#' values are inherently time-consuming, with the computational time increasing
-#' in proportion to the number of predictor variables and the number of observations.
-#' Therefore, the function uses a progress bar to check the progress of the kernel
-#' SHAP evaluation per observation.
+#' @details Shapley values (Shapley, 1953; Lundberg & Lee, 2017) apply the fair
+#' distribution of payoffs principles from game theory to measure the additive
+#' contribution of individual predictors in a ML model. The function compute
+#' a signed Shapley R2 metric, that combines the additive property of Shapley
+#' values with the robustness of the R-squared (R2) of Gelman (2018) to produce
+#' a variance decomposition that accurately captures the contribution
+#' of each variable in the ML model, see Redell (2019). The signed values are
+#' used in order to denote the regulation of connections in line with a linear
+#' model, since the edges in the DAG indicate node regulation (activation, if
+#' positive; inhibition, if negative). The sign has been recovered for each edge
+#' using sign(beta), i.e., the sign of the coefficient estimates from a linear
+#' model (lm) fitting of the output node on the input nodes (see Joseph, 2019).
+#' Furthermore, to determine the local significance of node regulation in the DAG,
+#' the Shapley decomposition of the R-squared values for each outcome node (r=1,...,R)
+#' can be done by summing the Shapley R2 indices of their input nodes.
+#' It should be noted that the operations required to compute Shapley values processed
+#' with the \code{\link[kernelshap]{kernelshap}} function of the \pkg{kernelshap} R
+#' package are inherently time-consuming, with the computational time increasing in
+#' proportion to the number of predictor variables and the number of observations.
 #'
 #' @return A list od four object: (i) shapx: the list of individual Shapley values
 #' of predictors variables per each response variable; (ii) est: a data.frame including
@@ -779,7 +689,7 @@ getVariableImportance <- function(object, thr = NULL, verbose = FALSE, ...)
 #' Red Hook, NY, USA, 4768–4777.
 #' 
 #' Redell, N. (2019). Shapley Decomposition of R-Squared in Machine Learning 
-#' Models. arXiv: Methodology.
+#' Models. arXiv preprint: https://doi.org/10.48550/arXiv.1908.09718
 #' 
 #' Gelman, A., Goodrich, B., Gabry, J., & Vehtari, A. (2019). R-squared for 
 #' Bayesian Regression Models. The American Statistician, 73(3), 307–309.
@@ -815,8 +725,9 @@ getVariableImportance <- function(object, thr = NULL, verbose = FALSE, ...)
 #'
 getShapleyR2<- function(object, newdata, newoutcome = NULL, thr = NULL, ncores = 2, verbose = FALSE, ...)
 {
+	stopifnot(inherits(object, c("SEM", "ML")))
+	
 	# extract model, formula and data objects
-	stopifnot(inherits(object, c("SEM", "DNN", "ML")))
 	dag <- object$graph
 	out <- unique(object$gest$lhs)
 	if (length(out) > 0) {
@@ -832,9 +743,6 @@ getShapleyR2<- function(object, newdata, newoutcome = NULL, thr = NULL, ncores =
 	 formula <- lapply(1:length(y), function(x){
 					C<- paste(y[[x]][,1], collapse = " + ")
 					f<- formula(paste(names(y)[x], "~", C))})
-	} else if (inherits(object, "DNN")){
-	 model <- object$model[[1]]
-	 formula <- lapply(1:length(model), function(x) model[[x]]$old_formula)
 	} else {
 	 model <- object$model[[1]]
 	 formula <- model$old_formula
@@ -858,11 +766,11 @@ getShapleyR2<- function(object, newdata, newoutcome = NULL, thr = NULL, ncores =
 	shapx <- list()
 	nrep <- length(formula)
 
-	#pb <- txtProgressBar(min = 0, max = length(model), style = 1)
-	pb <- pb(nrep = nrep, snow = FALSE)
+	#pb0 <- txtProgressBar(min = 0, max = length(model), style = 1)
+	pb0 <- pb(nrep = nrep, snow = FALSE)
 
 	for (j in 1:nrep) {
-	 pb$tick()
+	 pb0$tick()
 	 S <- shapR2(model[[j]], formula[[j]], Z_train, Z_test, ncores)
 	 W <- S$shapr2
 	 if (is.null(W)) {
@@ -879,9 +787,9 @@ getShapleyR2<- function(object, newdata, newoutcome = NULL, thr = NULL, ncores =
 		est <- rbind(est, cbind(label, W[vx,vy[k]]))
 	   }
 	 shapx<- c(shapx, list(S$shapx))
-	 #setTxtProgressBar(pb, j)
+	 #setTxtProgressBar(pb0, j)
 	}
-	#close(pb)
+	#close(pb0)
 	rownames(est) <- NULL
 	colnames(est)[4] <- "sign_r2"
 
@@ -917,13 +825,8 @@ shapR2<- function(fitj, fmj, Z_train, Z_test, ncores, ...)
 {
 	# Extract fitting objects
 	vf <- all.vars(fmj)
-	if (inherits(fitj, "citodnn")) {
-	 vx <- colnames(fitj$data$X)
-	 vy <- vf[vf %in% vx == FALSE]
-	} else {
-	 vx <- vf[-1]
-	 vy <- vf[1]
-	}
+	vx <- vf[-1]
+	vy <- vf[1]
 	z_train <- data.frame(Z_train[ ,vf])
 	z_test <- data.frame(Z_test[ ,vf])
 	if (nrow(z_train) < length(vx)) return(list(shapx = NULL, shapr2 = NULL))
@@ -980,16 +883,7 @@ shapR2<- function(fitj, fmj, Z_train, Z_test, ncores, ...)
 		predict(x, as.matrix(newdata[ ,x$feature_names]))
 	  }
  	}
-	if (inherits(fitj, "nnet")) {
-	  predict_model <- function(x, newdata) {
-		predict(x, newdata, type = "raw")
-	  }
-	}
-	if (inherits(fitj, "citodnn")) {
-	  predict_model <- function(x, newdata) {
-		predict(x, newdata, type = "response")
-	  }
- 	}
+
     #pos <- 1
 	envir <- as.environment(1)
 	assign("predict_model", predict_model, envir = envir)
@@ -1067,3 +961,210 @@ r2 <- function(shap, y, intercept, scale = c("r2", "1"))
 	
 	return(data)
 }
+
+#' @title Compute variable importance using LOCO values
+#'
+#' @description This function computes the contributions of each variable
+#' to individual predictions using LOCO (Leave Out COvariates) values.
+#'
+#' @param object A model fitting object from \code{SEMml()}, or \code{SEMrun()}
+#' functions.
+#' @param newdata A matrix containing new data with rows corresponding to subjects,
+#' and columns to variables.
+#' @param newoutcome A new character vector (as.factor) of labels for a categorical
+#' output (target)(default = NULL).
+#' @param thr A numeric value [0-1] indicating the threshold to apply to the
+#' LOCO values to color the graph. If thr = NULL (default), the
+#' threshold is set to thr = 0.5*max(abs(LOCO values)).
+#' @param ncores number of cpu cores (default = 2)
+#' @param verbose A logical value. If FALSE (default), the processed
+#' graph will not be plotted to screen.
+#' @param ... Currently ignored.
+#'
+#' @details LOCO (Verdinelli & Wasserman, 2024) is a model-agnostic method for
+#' assessing the importance of individual features (covariates) in a ML predictive
+#' model. The procedure is simple: (i) train a model on the full dataset (with all
+#' covariates) and (ii) for each covariate of interest: (a) remove (leave out) that
+#' covariate from the dataset; (b) retrain the model on the remaining features; (c)
+#' compare predictions between the full model and the reduced mode, and (d) evaluate
+#' the difference in performance (e.g., using MSE, etc.).
+#' LOCO is computationally expensive (requires retraining for each feature). The
+#' getLOCO() function uses a lowest computation cost procedure (see Delicando & Pena,
+#' 2023). The individual relevance of each variable is measured by comparing
+#' the predictions of the model in the test set with those obtained when the variable
+#' of interest is leave-out and substituted by its ghost variable in the test set.
+#' This ghost variable is defined as the linear prediction of the covariate by
+#' using the rest of the variables in the ML model. This method yields similar LOCO
+#' results but requires much less computing time.
+#'
+#' @return A list od three object: (i) est: a data.frame including the connections
+#' together with their LOCO values; (iii) gest: if the outcome vector is given,
+#' a data.frame of LOCO values per outcome levels; and (iii) dag: DAG with colored
+#' edges/nodes. If LOCO > thr, the edge is highlighted in red. If the outcome vector
+#' is given, nodes with absolute connection weights summed over the outcome levels,
+#' i.e. sum(LOCO[outcome levels]) > thr, will be highlighted in pink.
+#' 
+#' @author Mario Grassi \email{mario.grassi@unipv.it}
+#'
+#' @references
+#'
+#' Verdinelli, I; Wasserman, L. Feature Importance: A Closer Look at Shapley Values
+#' and LOCO. Statist. Sci. 39 (4) 623 - 636, November 2024. https://doi.org/10.1214/24-STS937
+#'
+#' Delicado, P.; Peña, D. Understanding complex predictive models with ghost variables.
+#' TEST 32, 107–145 (2023). https://doi.org/10.1007/s11749-022-00826-x
+#' 
+#' @examples
+#'
+#' \donttest{
+#' # load ALS data
+#' ig<- alsData$graph
+#' data<- alsData$exprs
+#' data<- transformData(data)$data
+#'
+#' #...with train-test (0.5-0.5) samples
+#' set.seed(123)
+#' train<- sample(1:nrow(data), 0.5*nrow(data))
+#'
+#' rf0<- SEMml(ig, data[train, ], algo="rf")
+#' 
+#' res<- getLOCO(rf0, data[-train, ], thr=0.2, verbose=TRUE)
+#' table(E(res$dag)$color)
+#' }
+#'
+#' @export
+#'
+getLOCO<- function(object, newdata, newoutcome = NULL, thr = NULL, ncores = 2, verbose = FALSE, ...)
+{
+	stopifnot(inherits(object, c("SEM", "ML")))
+		
+	# extract data, model, formula and data objects
+	dag <- object$graph
+	out <- unique(object$gest$lhs)
+	if (length(out) > 0) {
+	 dag <- mapGraph(dag, type="outcome", C=length(out))
+	 V(dag)$name[igraph::degree(dag, mode="out") == 0] <- out
+	}
+
+	if (inherits(object, "SEM")){
+	 object$data<- object$data[ ,-1]
+	 model <- list(object)
+	 fm <- NULL
+	} else if (inherits(object$model[[1]][[1]], "SEM")){
+	 model <- object$model
+	 fm <- NULL
+	} else {
+	 model <- object$model 
+	 fm <- model[[1]]$old_formula
+	}
+
+	if (!is.null(newoutcome)) {
+	 newout <- model.matrix(~newoutcome-1)
+	 colnames(newout) <- gsub("newoutcome", "", colnames(newout))
+	 newdata <- cbind(newout, newdata)
+	}
+	vp <- colnames(object$data)
+	mp <- apply(object$data, 2, mean)
+	sp <- apply(object$data, 2, sd)
+	Z_test <- scale(newdata[,vp], center=mp, scale=sp)
+	colnames(Z_test) <- paste0("z", "", colnames(Z_test))
+
+	#str(model, max.level=2)
+	nboot <- length(model)
+	est_boot <- list()
+	pb0 <- pb(nrep = nboot , snow = FALSE)
+
+	for(b in 1:nboot) {
+	  pb0$tick()
+	  	  
+	  est <- estLOCO(model[[b]], fm, Z_test, ncores)
+	  
+	  rownames(est) <- NULL
+	  colnames(est)[4] <- "loco"
+	  est_boot[[b]] <- est   
+	}
+	# str(est_boot, max.level=1)
+
+	EST <- do.call(cbind, lapply(est_boot, data.frame))
+	est <- cheapBootCI(EST[ ,c(1:3,seq(4, nboot*4, 4))], out, nboot)
+	dag0 <- colorDAG(dag, est, out, nboot, thr, verbose=verbose)
+
+	return(list(est = dag0$est, gest = dag0$gest, dag = dag0$dag))
+}
+
+estLOCO<- function(fit, fm, Z_test, ncores, ...)
+{
+	# extract LOCO for SEM
+	if (inherits(fit, "SEM")) {
+	 est <- parameterEstimates(fit$fit)
+	 est <- est[est$op == "~", c(1:3,4)]
+	 est[ ,4] <- est[ ,4]^2
+	 return(est)
+	}
+	if (inherits(fit[[1]], "SEM")) {
+	 est <- fit[[1]]$fit$parameterEstimates[ ,c(1:3,4)]
+	 est[ ,4] <- est[ ,4]^2
+	 return(est)
+	}
+
+	# start loop for ML
+	nrep <- length(fm)
+	est <- NULL
+
+	for (j in 1:nrep) {
+	  # extract fitting objects
+	  fitj <- fit[[j]]
+	  vf <- all.vars(fm[[j]])
+	  vx <- vf[-1]
+	  vy <- vf[1]
+	  Z <- data.frame(Z_test[ ,vf])
+	  X <- Z[ ,vx]
+	  Y <- Z[ ,vy]
+
+	  if (inherits(fitj, "ranger")){
+	   Yhat <- predict(fitj, Z, num.threads = ncores)$predictions
+	  } else if (inherits(fitj, "xgb.Booster")){
+	   Yhat <- predict(fitj, as.matrix(Z[,vx]))
+	  } else {
+	   Yhat <- predict(fitj, Z)
+	  }
+
+		D2 <- matrix(0, nrow=length(vx), ncol=length(vy))
+		for (k in 1:length(vx)) { #k=1
+		 Zk <- Z #; print(head(Zk))
+		 vk <- which(colnames(Zk) %in% vx[k])
+
+		 if (length(vx) > 1){
+		  C <- paste(vx[-k], collapse = " + ")
+		  f <- formula(paste(vx[k], "~", C)) #; print(f)
+		  Zk[ ,vk] <- predict(lm(eval(f), data = X)) #;print(head(Zk))
+		 } else {
+		  Zk[ ,vk] <- mean(Y) #;print(head(Zk))
+		 }
+
+		 if (inherits(fitj, "ranger")){
+		  Yhatk <- predict(fitj, Zk, num.threads = ncores)$predictions
+		 } else if (inherits(fitj, "xgb.Booster")){
+		  Yhatk <- predict(fitj, as.matrix(Zk[,vx]))
+		 } else {
+		  Yhatk <- predict(fitj, Zk)
+		 }
+
+		 D2[k,] <- colMeans((as.matrix(Yhat - Yhatk))^2)
+		} 
+		rownames(D2) <- vx <- sub(".", "", vx)
+		colnames(D2) <- vy <- sub(".", "", vy)
+		#print(D2); cat("\n")
+		 for (i in 1:length(vy)) {
+			esti <- data.frame(
+					  lhs = vy[i],
+					  op = "~",
+					  rhs = vx,
+					  est = D2[,i])
+			est <- rbind(est, esti)
+		 }			  
+	 #print(est); cat("\n")
+	}
+
+	return(est)
+}
